import logging

from contextlib import contextmanager
from datetime import datetime
from functools import wraps

from sqlalchemy import orm, create_engine, BigInteger, Column, DateTime, Integer, event, exc, String
from sqlalchemy.pool import Pool
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.ext.declarative import declarative_base, declared_attr
from sqlalchemy.types import TypeDecorator, SmallInteger

from dorthy.settings import config
from dorthy.request import RequestContextManager


logger = logging.getLogger(__name__)


class EntityBase(object):

    @declared_attr
    def __tablename__(cls):
        return cls.__name__.lower()

# Define Base class for ORM
Entity = declarative_base(cls=EntityBase)


class LookupMixin(object):

    native_code = Column(String(10), primary_key=True)
    friendly_name = Column(String(50), nullable=False)
    description = Column(String(1000), nullable=True)


class PrimaryKeyMixin(object):

    id = Column(BigInteger, primary_key=True)


class TimestampMixin(object):

    created = Column("created_ts", DateTime, nullable=False, default=datetime.today)


class UpdateTimestampMixin(TimestampMixin):

    updated = Column("updated_ts", DateTime)


class VersionedMixin(object):

    version_id = Column(Integer, nullable=False)

    __mapper_args__ = {
        "version_id_col": version_id
    }


class EnumIntType(TypeDecorator):

    impl = SmallInteger

    def __init__(self, values=None):
        super().__init__(self)
        self.values = values

    def process_bind_param(self, value, dialect):
        return None if value is None else self.values.index(value)

    def process_result_value(self, value, dialect):
        return None if value is None else self.values[value]


def get_enum_values(type_name):
    result = Session().execute("SELECT e.enumlabel " +
                               "FROM pg_enum e " +
                               "JOIN pg_type t ON e.enumtypid = t.oid " +
                               "WHERE t.typname=:typename", {"typename": type_name}).fetchall()
    return [row[0] for row in result]


@contextmanager
def transacted_session(**kwargs):

    session = Session()
    # currently only support new tx and joining existing
    session.begin(subtransactions=True, **kwargs)
    try:
        yield session
    except Exception:
        session.rollback()
        raise
    else:
        session.commit()


def transactional(*args):
    def wrap(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            with transacted_session():
                return f(*args, **kwargs)
        return wrapper
    return wrap


@transactional()
def add(entity):
    Session().add(entity)
    return entity


def find_by_id(entity_type, entity_id, not_found_error=True):
    try:
        return Session().query(entity_type).\
            filter(entity_type.id == entity_id).one()
    except NoResultFound:
        if not_found_error:
            raise
        return None


def find_all(entity_type):
    return Session().query(entity_type).all()


def _create_engine():
    # define DB Engine
    db_conf = config.db

    host = db_conf.host
    name = db_conf.name
    port = db_conf.port
    db_url = "postgresql://{user}:{password}@{host}:{port}/{db}".format(
        user=db_conf.username,
        password=db_conf.password,
        host=host,
        port=port,
        db=name)

    return create_engine(db_url,
                         pool_recycle=900,
                         echo="debug" if db_conf.enabled("debug") else False,
                         isolation_level="READ COMMITTED")


def _create_scoped_session(scope_func=None):
    # autocommit true is used for begin / commit functionality in
    # transactional decorators -- subtransaction support
    # http://docs.sqlalchemy.org/en/rel_0_7/orm/session.html#session-subtransactions
    return orm.scoped_session(orm.sessionmaker(bind=_create_engine(), autocommit=True), scope_func)


class SessionContext(object):

    def __init__(self):
        self.__request_context_scope = \
            _create_scoped_session(scope_func=SessionContext._request_context_scope_func)
        self.__thread_local_scope = \
            _create_scoped_session()

    def __call__(self, *args, **kwargs):
        if self.request_context_scoped:
            session = self.__request_context_scope()
        else:
            session = self.__thread_local_scope()
        return SessionContext._init_commit_callback(session)

    @property
    def request_context_scoped(self):
        return RequestContextManager.active()

    def remove(self):
        if self.request_context_scoped:
            self.__request_context_scope.remove()
        else:
            self.__thread_local_scope.remove()

    @staticmethod
    def _init_commit_callback(session):
        if not hasattr(session, "register_after_commit"):
            session.after_commit_callbacks = []
            session.register_after_commit = lambda cb, handle_error=False: \
                session.after_commit_callbacks.append((cb, handle_error))
        return session

    @staticmethod
    def _request_context_scope_func():
        request_context = RequestContextManager.get_context()
        if not request_context.contains_listener(SessionContext._request_context_session_release):
            request_context.register_listener(SessionContext._request_context_session_release)
        return request_context.id

    @staticmethod
    def _request_context_session_release(event, context):
        # remove sessions on deactivate events
        if event == "deactivate":
            try:
                Session.remove()
            except:
                logger.warn("Failed to remove db session from request context: %s", context.id)


# Global session for the application
# Sessions are scoped - request or thread local
Session = SessionContext()


#
# Configure Event Listeners
#
def update_timestamp(mapper, connection, target):
    target.updated = datetime.today()


def validator(mapper, connection, target):
    if hasattr(target, "validate"):
        target.validate()


@event.listens_for(orm.mapper, "mapper_configured")
def set_events(mapper, cls):
    if issubclass(cls, UpdateTimestampMixin):
        event.listen(cls, "before_update", update_timestamp)
    if hasattr(cls, "validate"):
        event.listen(cls, "before_update", validator)
        event.listen(cls, "before_insert", validator)


@event.listens_for(Pool, "checkout")
def ping_connection(dbapi_connection, connection_record, connection_proxy):
    cursor = dbapi_connection.cursor()
    try:
        cursor.execute("SELECT 1")
    except:
        # optional - dispose the whole pool
        # instead of invalidating one at a time
        # connection_proxy._pool.dispose()

        # raise DisconnectionError - pool will try
        # connecting again up to three times before raising.
        raise exc.DisconnectionError()
    else:
        cursor.close()


@event.listens_for(orm.Session, "after_commit")
def exec_commit_callbacks(session):
    if hasattr(session, "after_commit_callbacks") and session.after_commit_callbacks:
        try:
            for cb in session.after_commit_callbacks:
                try:
                    cb[0]()
                except:
                    if not cb[1]:
                        raise
                    logger.exception("Failed to process after commit callback.")
        finally:
            session.after_commit_callbacks[:] = []


import datetime
import hashlib
import logging
import sys

from .core import AuthenticationException
from .crypto import sign_message

from dorthy.utils import rfc_2822_timestamp, convert_from_rfc_2822, now, xstr


logger = logging.getLogger(__name__)


class SignedRequestAuthenticationToken(object):

    def __init__(self, token, request):
        self.token = token
        self.request = request

    @property
    def access_id(self):
        return self.token.split(":")[0]

    @property
    def signature(self):
        return self.token.split(":")[1]


class AbstractSignedRequestAuthenticationProvider(object):
    """This provides a signature request based authentication provider.
    """

    def __init__(self, ttl=120):
        self.__ttl = ttl

    def supports(self, authentication_token):
        if authentication_token and isinstance(authentication_token, SignedRequestAuthenticationToken):
            return True
        else:
            return False

    def authenticate(self, authentication_token):

        if not verify_request_ttl(authentication_token.request, self.__ttl):
            raise AuthenticationException("Request expired.")

        if not verify_http_request(authentication_token.request,
                                   authentication_token.signature,
                                   self._get_signing_key(authentication_token)):
            raise AuthenticationException("Invalid signature.")

        self._authenticate_user(authentication_token)

    def _authenticate_user(self, authentication_token):
        raise NotImplementedError

    def _get_signing_key(self, authentication_token):
        raise NotImplementedError


def sign_http_request(verb, uri_path, timestamp, signing_key, content_md5=None, query=None, encoding=None):
    if not encoding:
        encoding = sys.getdefaultencoding()
    if not isinstance(timestamp, str):
        timestamp = rfc_2822_timestamp(timestamp)
    if not isinstance(signing_key, bytes):
        signing_key = signing_key.encode(encoding)
    if isinstance(content_md5, bytes):
        content_md5 = content_md5.decode(encoding)
    msg = "{}\n{}\n{}\n{}\n{}".format(
        verb, uri_path, xstr(content_md5), xstr(query), timestamp).encode(encoding)
    return sign_message(signing_key, msg).decode(encoding)


def verify_http_request(request, request_signature, signing_key, encoding=None):
    if not encoding:
        encoding = sys.getdefaultencoding()
    if "Date" not in request.headers:
        raise AuthenticationException("No timestamp found in request.")
    timestamp = request.headers["Date"]
    content_md5 = None
    if request.body:
        content_md5 = hashlib.md5(request.body).hexdigest()
    signed_request = sign_http_request(request.method,
                                       request.path,
                                       timestamp,
                                       signing_key,
                                       content_md5=content_md5,
                                       query=request.query,
                                       encoding=encoding)
    return signed_request == request_signature


def verify_request_ttl(request, ttl=120):
    if "Date" not in request.headers:
        raise AuthenticationException("No timestamp found in request.")
    timestamp = convert_from_rfc_2822(request.headers["Date"])
    return timestamp + datetime.timedelta(seconds=ttl) >= now()


from distutils.core import setup
from setuptools import find_packages

version = "0.4.5"

install_requires = [
    "PyYAML",
    "tornado",
    "redis",
    "raven",
    "pycrypto",
    "SQLAlchemy",
    "jinja2",
    "py3k-bcrypt",
    "pytz"
]

setup(
    name='dorthy',
    version=version,
    packages=find_packages(),
    url='https://github.com/MobileDynasty/dorthy',
    author='dev@mobile-dynasty.com',
    author_email='dev@mobile-dynasty.com',
    license='MIT',
    description="a micro web framework for Tornado",
    long_description='a micro web framework for Tornado',
    keywords='web framework development',
    install_requires=install_requires,
    zip_safe=False,
    classifiers=[
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        'Intended Audience :: Developers',
        'Topic :: Internet :: WWW/HTTP',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4'
    ]
)
